/*
 * arff generator class
 */

#include "arff.h"
#include <assert.h>
#include <err.h>

void arff::add_comment(string comment)
{
    comments.push_back(comment);
}

void arff::add_attribute(string name,int code)
{
    /* Don't add if already there */
    for(int i=0;i<attributeTypes.size();i++){
	if(attributeNames[i]==name){
	    if(attributeTypes[i]!=code) err(1,"attribute %s changed type\n",name.c_str());
	    return;
	}
    }
    attributeTypes.push_back(code);
    attributeNames.push_back(name);
    attributeCodes[name] = attributeNames.size()-1;
}

void arff::new_row()
{
    values.push_back(new valueMapT);
}

void arff::add_value(string name,int value)
{
    int code = attributeCodes[name];
    char buf[64];
    sprintf(buf,"%d",value);
    (*values.back())[code] = buf;
}

void arff::add_valuet(string name,time_t t)
{
    int code = attributeCodes[name];
    char buf[64];
    strftime(buf,sizeof(buf),"%F %T",gmtime(&t));
    (*values.back())[code] = buf;
}

void arff::add_value(string name,const string &value)
{
    int code = attributeCodes[name];
    (*values.back())[code] = value;
}

static bool has_non_numeric(const string &s)
{
    for(int i=0;i<s.size();i++){
	if(isdigit(s[i])==0 && s[i]!='.' && s[i]!='-' && s[i]!='+') return true; // it has a non-numeric
	if(i>0 && s[i]=='-' ) return true;
	if(i>0 && s[i]=='+' ) return true;
    }
    return false;
}

static bool is_date(const string &s)
{
    if(isdigit(s[0]) && isdigit(s[1]) && isdigit(s[2]) && isdigit(s[3]) && s[4]=='-' &&
       isdigit(s[5]) && isdigit(s[6]) && s[7]=='-' &&
       isdigit(s[8]) && isdigit(s[9])  &&
       s[10]==' ' &&
       isdigit(s[11]) && isdigit(s[12]) && s[13]==':' &&       
       isdigit(s[14]) && isdigit(s[15]) && s[16]==':' &&       
       isdigit(s[17]) && isdigit(s[17]) && s.size()==19) return true;
    return false;
}

int arff::attributeCol(const string &attributeName) const
{
    for(int i=0;i<attributeNames.size();i++){
	if(attributeNames[i]==attributeName) return i;
    }
    return -1;
}

/* Scan the values we have for an attribute and return true if it always missing or numeric. */
bool arff::attributeAlwaysNumeric(const string &attributeName) const
{
    int col = attributeCol(attributeName);
    for(vector<valueMapT *>::const_iterator row = values.begin(); row!= values.end(); row++){
	valueMapT::const_iterator k = (*row)->find(col); // find the row we are looking for
	if(k==(*row)->end()) continue;		   // no value
	const string &val = (*k).second.c_str();	   // no need to copy
	if(has_non_numeric(val)) return false;
    }
    return true;
}

bool arff::attributeAlwaysDate(const string &attributeName) const
{
    int col = attributeCol(attributeName);
    for(vector<valueMapT *>::const_iterator row = values.begin(); row!= values.end(); row++){
	valueMapT::const_iterator k = (*row)->find(col); // find the row we are looking for
	if(k==(*row)->end()) continue;		   // no value
	const string &val = (*k).second.c_str();	   // no need to copy
	if(is_date(val)==false) return false;
    }
    return true;
}

/** return true if a string needs quoting */
static bool needs_quoting(const string &str){
    for(int i=0;i<str.size();i++){
	switch(i){
	case ' ':
	case '{':
	case '}':
	case ',':
	case '\t':
	    return true;
	}
    }
    return false;
}

void arff::output(FILE *o)
{
    /* Before we generate the output, see if any of the string attributes always have numeric values.
     * If so, we'll change them into numeric values...
     */
    for(int i=0;i<attributeNames.size();i++){
	if(attributeTypes[i]==STRING && attributeAlwaysNumeric(attributeNames[i])){
	    attributeTypes[i]=NUMERIC;
	}
	if(attributeTypes[i]==STRING && attributeAlwaysDate(attributeNames[i])){
	    attributeTypes[i]=DATE;
	}
    }


    for(vector<string>::const_iterator i = comments.begin();
	i!= comments.end();
	i++){
	fprintf(o,"%% %s\n",i->c_str());
    }
    fprintf(o,"%\n");
    fprintf(o,"@RELATION %s\n",relation.c_str());
    fprintf(o,"\n");
    
    for(int i=0;i<attributeNames.size();i++){
	string name = attributeNames[i];
	switch(attributeTypes[i]){
	case NUMERIC:
	    fprintf(o,"@ATTRIBUTE %s NUMERIC\n",name.c_str());
	    break;
	case DATE:
	    fprintf(o,"@ATTRIBUTE %s date \"yyyy-MM-dd HH:mm:ss\"\n",name.c_str());
	    break;
	case STRING:
	    fprintf(o,"@ATTRIBUTE %s string\n",name.c_str());
	    break;
	default:
	    assert(0);
	}
    }
    fprintf(o,"\n@DATA\n\n");
    for(vector<valueMapT *>::const_iterator row = values.begin();
	row!= values.end();
	row++){
	for(int i=0;i<attributeNames.size();i++){
	    valueMapT::const_iterator k = (*row)->find(i); // find the row we are looking for
	    string val("?");		// default value
	    if(k!=(*row)->end()){
		val = (*k).second.c_str();
		/* Remove any quotes */
		for(int j=0;j<val.size();j++){
		    if(val[j]=='"') val[j]=' ';
		    if(val[j]=='\'') val[j]=' ';
		}
		/* If there are any spaces quote the result */
		if(needs_quoting(val)){
		    val = "\"" + val + "\"";
		}
	    }
	    if(val.size()==0) val="?";	// put back the ?
	    if(i>0) fprintf(o,", ");
	    fprintf(o,"%s",val.c_str());
	}
	fprintf(o,"\n");
    }
    fflush(o);
}

#ifdef STANDALONE
main()
{
    printf("ARFF Test Program\n");
    arff a("test relation");

    a.add_comment("This is a test file");
    a.add_comment("Just a test");
    a.add_attribute("blue",arff::NUMERIC);
    a.add_attribute("green",arff::NUMERIC);
    a.new_row();
    a.add_value("blue",1);
    a.add_value("green",10);
    a.new_row();
    a.add_value("green",20);
    a.add_value("blue",2);
    a.new_row();
    a.add_value("blue",3);
    a.output(stdout);
}
#endif
