/*
 * fiwalk.cpp:
 * File and Inode Walk.
 *
 * This application uses SleuthKit to generate a report of all of the files
 * and orphaned inodes found in a disk image. It can optionally compute the MD5 of 
 * any objects, save those objects into a directory, or both.
 *
 * Algorithm:
 * 1 - Find all of the partitions on the disk. 
 * 2 - For each partition, walk the files.
 * 3 - For each file, print the requested information.
 * 4 - For each partition, walk the indoes
 * 5 - For each inode, print the requested information.
 */


#include <stdio.h>
#include <strings.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/time.h>
#include <ctype.h>

#include <openssl/md5.h>

/* Bring in our headers */
#include "hexbuf.h"
#include "arff.h"
#include "plugin.h"

#include <err.h>

/* Bring in TSK headers */
//#ifdef linux
//#define LINUX2
//#endif

//#include "tsk_os.h"
//#include "aux_tools.h"
//#include "fs_tools.h"
//#include "mm_tools.h"
#include <tsk/libtsk.h>

/* Bring in AFFLIB headers */
#include <afflib/afflib.h>

#include <algorithm>
#include <cstdlib>
#include <vector>
#include <string>

/* Structure for keeping track of file segments */
struct seg {
    int64_t start;			// byte offset from start of image
    int64_t len;				// number of bytes
};

/* Configurable options */
char *iwalk_version = "0.3";
int  do_orphan=0;
int  do_allocated=0;
char *save_outdir = "";
int  opt_debug    = 0;
int  opt_md5      = 0;			// calculate the MD5 for every file?
bool opt_save     = false;
bool do_plugin = false;		// this one requires plugin processing
const char *config_file = 0;

int  out_arff = 0;
int  out_xml  = 0;
int  out_text = 0;
int  row_number   = 0;

class arff *a=0;			// ARFF generator

using namespace std;
typedef vector<struct seg> segs_;	// vector of blocks
segs_ segs;

vector<string> namelist;		// names of files that we want to find

/****************************************************************
 ** Current information.
 **
 ****************************************************************/
int    current_partition_num=0;
int64_t  current_partition_start=0;	// in bytes
string current_filename;		// filename of what's currently being saved
string current_path;	 // full path of file currently being processed
string save_path;	// full path of where it is being saved

/* Individual 'state' variables */
MD5_CTX md5;
FILE   *outfile = 0;
FILE   *plugin_outfile = 0;		// where to send for plugin processing
string plugin_filename;

void usage()
{
    printf("usage: iwalk [options] iso-name\n");
    printf("Default behavior: Just print the file system statistics and exit.\n");
    printf("options:\n");
    printf("    -F = report allocated files\n");
    printf("    -O = report orphan files  (default is \n");
    printf("    note: if neither -F nor -O is provided, -FO is assumed\n");
    printf("    -c config.txt = read config.txt for metadata extraction tools\n");
    printf("    -n name  = only match files that contain 'name'. May be repeated for OR\n");
    printf("              case of 'name' is ignored.\n");
    printf("              example: -n jpeg -n jpg will find all JPEG files\n");
    printf("              (except those that are recovered by carving)\n");
    printf("\n");
    printf("\nFull content options (slow processing significantly):\n");
    printf("    -m = Also report MD5 for each file\n");
    printf("    -s <dir> = Save all recovered files to <dir>\n");
    printf("\n");
    printf("Output options:\n");
    printf("    -A = ARFF output\n");
    printf("    -X = XML output (not currently implemented)\n");
    printf("\n");
    printf("    -d = debug this program\n");
    printf("    -v = Enable SleuthKit verbose flag\n");
    printf("\n");
    exit(1);
}



/* Note: TSK has a broken MD5 implementation, but we can't define a better
 * one because of some namespace conflicts. So just use this:
 */

//#define MD5_Update MD5Update
//#define MD5_Init   MD5Init
//#define MD5_Final  MD5Final


/****************************************************************
 ** Output
 ****************************************************************/

/**
 * output a comment in the current file format
 */

void comment(const char *format,...)
{
    va_list ap;
    if(out_arff) fputs("% ",stdout);
    if(out_xml)  fputs("<!-- ",stdout);
    va_start(ap, format);
    (void) vprintf(format,ap);
    va_end(ap);
    if(out_xml)  fputs("--> ",stdout);
    putchar('\n');
}

/* TTD: Modify Block Runs so it prints NN-NN if block is length 1, or change
 * output format to Block[+LEN]
 */

/* open a save filename. If the file exists, then start a counter
 * between the end of the file and the extension and keep incrementing...
 */
FILE *fopen_savefile(const char *filename)
{
    if(access(save_path.c_str(),F_OK)!=0){
	return fopen(save_path.c_str(),"w");
    }

    char fmt[1024];
    const char *ext = strrchr(filename,'.');

    strcpy(fmt,filename);
    char *cc = strchr(fmt,'.');
    if(cc) *cc = '\000';
    strcat(fmt,".%d");
    if(ext) strcat(fmt,ext);

    for(int i=1;i<10000;i++){
	char fname[1024];
	snprintf(fname,sizeof(fname),fmt,i);
	if(access(fname,F_OK)!=0){
	    return fopen(fname,"w");
	}
    }
    return 0;				// hm. Can't open file?
}



/* block_map (will) keeps track of all the blocks that have been touched
 * by the regular walking, so that it can be carved.
 * This isn't implemented right now. 
 */
unsigned char *block_map = 0;

/****************************************************************
 *** Support Code.
 ****************************************************************/

/* FS_INFO defined in ~/sleuthkit-2.05/src/fstools/fs_tools.h
 * IMG_INFO defined in img_tools.h 
 * img_open - opens the image; calls do_dimage
 * do_dimage - analyizes the image
 *        mm_open - opens the volume; mm_part_walk - walk each volume;
 *                  calls mm_act for each
 * mm_act - analyizes each partition; calls do_vol for ones that we can analyze
 * do_vol - analyizes each volume
 *       fs_open - opens the file system; dent_walk 
 *               - walks the directory structure.
 *               - calls dent_act for each file
 * dent_act - prints file name; calls file_walk() to read each file; 
 *          - calls file_act for each sector
 * file_act - prints the block number
 */

/* 
 * dent_walk callback function
 *  -- obtain the address of each data unit
 *  -- optionally calculate the mac or save the data
 */
TSK_WALK_RET_ENUM
file_act(TSK_FS_INFO * fs, TSK_DADDR_T addr, char *buf, size_t size, TSK_FS_BLOCK_FLAG_ENUM flags, void *ptr)
{
    //printf("file_act: addr=%qd buf=%x size=%d flags=%x\n",addr,buf,size,flags);

    if(opt_debug){
	printf("addr=%"PRIuDADDR" buf=%p size=%d\n",addr,buf,(int)size);
	if(opt_debug>1 && segs.size()==0){
	    fwrite(buf,size,1,stdout);
	    printf("\n");
	}
    }

    if(size==0)  return TSK_WALK_CONT;	    // can't do much with this...
    if(opt_md5)  MD5_Update(&md5,(unsigned char *)buf,size);
    if(opt_save) fwrite(buf,size,1,outfile);
    if(plugin_outfile) fwrite(buf,size,1,plugin_outfile);

    /* See if we can extend the last segment in the segment list,
     * or if this is the start of a new fragment.
     */
    int64_t byteaddr = current_partition_start + ((int64_t)addr)*fs->block_size;

    if(segs.size()>0){
	if(segs.back().start + segs.back().len == (int64_t)byteaddr){
	    segs.back().len += size;
	    return TSK_WALK_CONT;
	}
    }
    /* Need to add a new element to the list */
    struct seg newseg;
    newseg.start = byteaddr;
    newseg.len   = size;
    segs.push_back(newseg);
    return TSK_WALK_CONT;
}


/*
 * inode_act:
 * Called for each file in the file system,
 * whether it is a named file or an orphan file. 
 * (ptr==NULL for orphans).
 *
 * Walk each file to print the addresses
 */
static TSK_WALK_RET_ENUM
inode_act(TSK_FS_INFO * fs, TSK_FS_INODE * fsi, void *ptr)
{
    unsigned char md5buf[16];

    /* If the name has corresponding metadata, then walk it */
    if(!fsi) return TSK_WALK_CONT;	// no file system info??

    if ((fsi->mode & TSK_FS_INODE_MODE_FMT)!=TSK_FS_INODE_MODE_REG){
	/* Only do regular files */
	return TSK_WALK_CONT;
    }

    /* If we are filtering and we have a filename, see if we want this file. */
    if(namelist.size()>0 && current_filename.size()>0){
	bool file_wanted = false;
	for(vector<string>::const_iterator i = namelist.begin(); i!=namelist.end(); i++){
	    if(strcasestr(current_filename.c_str(),i->c_str())){
		file_wanted = true;
	    }
	}
	if(!file_wanted) return TSK_WALK_CONT;
    }

    /* Looks like we are processing */
    if(a){
	a->new_row();
	a->add_value("id",++row_number);
    }
	    
    /* See if this one requires plugin processing */
    do_plugin = plugin_match(current_filename.c_str());
    if(do_plugin){
	plugin_filename = current_filename;
	/* Remove special characters */
	for(int i=0;i<plugin_filename.size();i++){
	    if(isalnum(plugin_filename[i])==0) plugin_filename[i]='_';
	}
	plugin_filename = "/tmp/" + plugin_filename;
	plugin_outfile = fopen(plugin_filename.c_str(),"w");
	if(!plugin_outfile) errx(1,plugin_filename.c_str());
    }

    int full_content = (opt_md5) || (opt_save) || (do_plugin); // do we want fully content?
    segs.clear();				// segments of where the file is

    int myflags = TSK_FS_FILE_FLAG_NOID;
    myflags |= TSK_FS_FILE_FLAG_RECOVER; // use special data recovery for deleted files

    if(opt_md5) 	MD5_Init(&md5);	// turn on MD5 machinery if desired
    if(opt_save)	outfile = fopen_savefile(save_path.c_str());

    if(!full_content) {
	myflags |= TSK_FS_FILE_FLAG_AONLY;	/* Just get the metadata */
    }
	
    /* Now get the information on the individual
     * sectors that make up the file.
     */
    if (fs->file_walk(fs, fsi, 0, 0, (TSK_FS_FILE_FLAG_ENUM)myflags, file_act, NULL)) {
	tsk_error_reset();
    }
    if(opt_md5){			// finish up the MD5 calc
	MD5_Final(md5buf,&md5);
    }
    if(opt_save){
	fclose(outfile);
	outfile = 0;
    }
	
    /* Send through to the plugin if we were doing that. */
    if(do_plugin){
	fclose(plugin_outfile);
	plugin_outfile = 0;
	plugin_process(plugin_filename.c_str(),a);
	//unlink(plugin_filename.c_str());
    }

    /* Finally output the informaton */
    if(out_text){
	if(current_path.size()>0) printf("Filename: %s\n",current_path.c_str());
	if(ptr==NULL) printf("Orphan: YES\n");
	printf("Filesize: %qd\n",(long long)fsi->size);
	printf("Status: ");
	if(fsi->flags & TSK_FS_INODE_FLAG_ALLOC)   printf("ALLOC ");
	if(fsi->flags & TSK_FS_INODE_FLAG_UNALLOC) printf("UNALLOC ");
	if(fsi->flags & TSK_FS_INODE_FLAG_USED)    printf("USED ");
	if(fsi->flags & TSK_FS_INODE_FLAG_UNUSED)  printf("UNUSED ");
	if(fsi->flags & TSK_FS_INODE_FLAG_ORPHAN)  printf("ORPHAN ");
	printf("\n");
	if(fsi->mtime) printf("mtime: (%d) %s", (int)fsi->mtime, ctime(&fsi->mtime));
	if(fsi->ctime) printf("ctime: (%d) %s", (int)fsi->ctime, ctime(&fsi->ctime));
	if(fsi->atime) printf("atime: (%d) %s", (int)fsi->atime, ctime(&fsi->atime));
	if(segs.size()>0){
	    printf("Byte runs: ");
	    for(segs_::const_iterator i = segs.begin();i!=segs.end();i++){
		printf("%qd:%qd ",i->start,i->len);
	    }
	    printf("\n");
	}
	
	printf("Fragments: %d\n",(int)segs.size());
	if(opt_md5){
	    char buf[64];
	    hexbuf(buf,sizeof(buf),md5buf,16,0);
	    printf("MD5: %s\n",buf);
	}
	printf("\n");
	fflush(stdout);
    }
    if(a){
	a->add_value("partition",current_partition_num);
	a->add_value("filesize",fsi->size);
	a->add_valuet("mtime",fsi->mtime);
	a->add_valuet("ctime",fsi->ctime);
	a->add_valuet("atime",fsi->atime);
	a->add_value("fragments",(int)segs.size());
	if(fs->block_size>0){
	    if(segs.size()>=1) a->add_value("frag1startsector",(int)(segs[0].start / fs->block_size));
	    if(segs.size()>=2) a->add_value("frag2startsector",(int)(segs[1].start / fs->block_size));
	}
	if(current_path.size()>0) a->add_value("filename",current_path);
    }

    return TSK_WALK_CONT;
}


/*
 * dent_acct:
 * Called for each file in the file system
 * Callback action for the file system dent_walk
 *
 * Walk each file to print the addresses
 */
TSK_WALK_RET_ENUM
dent_act(TSK_FS_INFO * fs, TSK_FS_DENT * fs_dent, void *ptr)
{
    current_filename = "";
    current_path = "";
    save_path = "";
    if (fs_dent->fsi) {
	current_filename = fs_dent->name;
	current_path = string(fs_dent->path) + current_filename;
	if(opt_save) save_path = string(save_outdir) + "/" + current_filename;
	return inode_act(fs,fs_dent->fsi,ptr);
    }
    return TSK_WALK_CONT;
}


/* 
 * Analyze the volume starting at byte offset 'start' in the image
 * file pointed to by img. 
 *
 * Return 0 on success and 1 on error 
 */
int do_vol(TSK_IMG_INFO *img, off_t start)
{
    TSK_FS_INFO *fs;
    int ret = 0;

    if (tsk_verbose) fprintf(stderr, "do_vol:  Start: %" PRIuDADDR "\n", start);

    /* Try it as a file system */
    if ((fs = tsk_fs_open(img, start, NULL)) == NULL) {
	tsk_error_print(stderr);

	/* @@@ We could do some file system carving at this point */
	return 1;
    }

    current_partition_num++;
    current_partition_start = fs->offset;
    if(out_text){
	printf("Partition Offset: %"PRIuDADDR" (sector %"PRIuDADDR") ", fs->offset,fs->offset/512);
	printf("block_size=%d  ftype=%d  (%s) ",fs->block_size,fs->ftype,tsk_fs_get_type(fs->ftype));
	printf("block_count=%"PRIuDADDR"  ",fs->block_count);
	printf("first_block=%"PRIuDADDR"  ",fs->first_block);
	printf("last_block: %"PRIuDADDR"  ",fs->last_block);
	printf("\n");
	printf("\n");
	fflush(stdout);
    }

    block_map = (unsigned char *)calloc(1,fs->last_block);

    /*
     * If we didn't want information on the files, we could just
     * return here with
     *
     * fs->close(fs); return 0;
     */
    
    /* Walk the files, starting at the root directory */
    if(do_allocated){
	time_t t = time(0);
	comment("Starting dent_walk at %s",ctime(&t));
	int dent_walk_flags = (TSK_FS_DENT_FLAG_ALLOC | TSK_FS_DENT_FLAG_UNALLOC |
			       TSK_FS_DENT_FLAG_RECURSE);
	if (fs->dent_walk(fs, fs->root_inum, (TSK_FS_DENT_FLAG_ENUM)dent_walk_flags, dent_act, &ret)) {
	    tsk_error_print(stderr);
	    ret = 1;
	}
    }

    if(do_orphan){
	time_t t = time(0);
	comment("Performing orphan inode_walk from %qd to %qd at %s",
		(long long)fs->first_inum,(long long)fs->last_inum,ctime(&t));

	/* Don't need other flags; it will fill in the rest */
	TSK_FS_INODE_FLAG_ENUM inode_walk_flags = TSK_FS_INODE_FLAG_ORPHAN; 
	if (fs->inode_walk(fs,fs->first_inum,fs->last_inum,inode_walk_flags, inode_act, NULL)){
	    tsk_error_print(stderr);
	    ret = 1;
	}
    }

    fs->close(fs);
    free(block_map);
    return ret;
}

/*
 * Callback for the partition table walk.
 *
 * Analyze the file system for each partition.
 *
 */
TSK_WALK_RET_ENUM mm_act(TSK_MM_INFO * mm, TSK_PNUM_T pnum, TSK_MM_PART * part, int flag, void *ptr)
{
    // Currently, skip the "descriptive" entries
    if (part->type == TSK_MM_PART_TYPE_DESC){
	return TSK_WALK_CONT;
    }

    if (do_vol(mm->img_info, part->start * mm->block_size)) {
	return TSK_WALK_ERROR;
    }
    return TSK_WALK_CONT;
}



/* do_dimage:
 * Process the partition tables in the disk image pointed to by img.
 * File system analysis will be performed on each partition.
 *   mm_open() - opens partition
 *   mm_part_walk() - walks partition
 * If we can't open as a disk image, try to open as a raw file system
 *
 * returns 1 on error and 0 on success
 */
int do_dimage(TSK_IMG_INFO * img, char *desc)
{
    TSK_MM_INFO *mm;

    if ((mm = tsk_mm_open(img, 0, NULL)) == NULL) {
	if (tsk_verbose) fprintf(stderr, "Error determining volume system -- trying file systems\n");

	/* Check if we have a partition image */
	tsk_error_reset();
	if (do_vol(img, 0)) {
	    return 1;
	}
    }
    else {
	if (tsk_verbose) fprintf(stderr, "Volume system open, examining each\n");
	/* Walk the volumes and update the volume table */
	if (mm->part_walk(mm, mm->first_part, mm->last_part, 0, mm_act, 0)) {
	    return 1;
	}
    }
    return 0;
}


int main(int argc, char *argv[])
{
    TSK_IMG_INFO *img;
    int ch;
    extern int optind;

    int t0 = time(0);

    while ((ch = getopt(argc, argv, "mdvOFs:AXn:c:")) > 0) {
	switch (ch) {
	case 'm':
	    opt_md5++;
	    break;
	case 'd':
	    opt_debug++;
	    break;
	case 'v':
	    tsk_verbose++;			// sleuthkit option
	    break;
	case 'F': do_orphan = 1; break;
	case 'O': do_allocated=1; break;
	case 's': save_outdir = optarg; opt_save = true; break;
	case 'A': out_arff = 1; break;
	case 'X': out_xml = 1;break;
	case 'n': namelist.push_back(optarg);break;
	case 'c': config_file = optarg; break;
	case 'h':
	case '?':
	default:
	    fprintf(stderr, "Invalid argument: %s\n", argv[optind]);
	    usage();
	}
    }

    /* If neither -O and -F are specified, do both! */
    if(do_orphan==0 && do_allocated==0){
	do_orphan = 1;
	do_allocated = 1;
    }

    if (optind >= argc) usage();
    argc -= optind;
    argv += optind;

    if(config_file) config_read(config_file);    /* Read the configuration file */

    /* Check that we have a valid file format */
    time_t t = time(0);
    comment("version: %s",iwalk_version);
    comment("start time: %s",ctime(&t));

    const char *filename = argv[0];
    if ((img = tsk_img_open(NULL, argc, (const char **)argv)) == NULL) {
	tsk_error_print(stderr);
	comment("=EOF= (img_open)");
	exit(1);
    }
    comment("image filename: %s",filename);

    if(out_arff==0 && out_xml==0){
	out_text = 1;
    }

    /* Set up the output */
    if(out_arff){
	a = new arff("fiwalk");		// the ARFF output object
	a->add_attribute("id",arff::NUMERIC);
	a->add_attribute("partition",arff::NUMERIC);
	a->add_attribute("filesize",arff::NUMERIC);
	a->add_attribute("mtime",arff::DATE);
	a->add_attribute("ctime",arff::DATE);
	a->add_attribute("atime",arff::DATE);
	a->add_attribute("fragments",arff::NUMERIC);
	a->add_attribute("frag1startsector",arff::NUMERIC);
	a->add_attribute("frag2startsector",arff::NUMERIC);
	a->add_attribute("filename",arff::STRING);
    }

    if (out_text) printf("Imagefile: %s\n",filename);
    if (do_dimage(img, "my boring test comment")) {
	tsk_error_print(stderr);
	comment("=EOF= (do_dimage)");
	exit(1);
    }
    img->close(img);

#ifdef HAVE_GETRUSAGE
    /* Print usage information */
    struct rusage ru;
    memset(&ru,0,sizeof(ru));
    int t1 = time(0);
    printf("# clock: %d\n",t1-t0);
    if(getrusage(RUSAGE_SELF,&ru)==0){
	t = time(0);
	comment("u: %d  s: %d\n",(int)ru.ru_utime.tv_sec,(int)ru.ru_stime.tv_sec);
	comment("maxrss: %ld reclaims: %ld faults: %ld   swaps: %ld\n",
		ru.ru_maxrss,ru.ru_minflt,ru.ru_majflt,ru.ru_nswap);
	comment("inputs: %d  outputs: %d\n",(int)ru.ru_inblock,(int)ru.ru_oublock);
	comment("stop time: %s",ctime(&t));
    }
#endif
    if(out_arff) a->output(stdout);
    comment("=EOF=");
    exit(0);
}

